/*
30-June-2005  Drew Baden
1. Start with s2 version of the sandwich self-triggering board....
2. Modify so that each Xilinx in the HTR sends, at 80MHz
    o 2 bits per channel (24 channels, 48 bits)
	 o 16 bits for sumET
	 o 1 bit for "frame"
	 o 1 bit for L1A
  Total: 66 bits @ 80MHz
3. We need some output pins to debug, so we'll use the "Wesley" connectors:
  J4 and J5 are the 50pin connectors that are available and look like this
  on the board:
  
     J4          J5         J7
  -50- -25-   -50- -25-    |||||
  -49- -24-   -49- -24-    |||||
  
  -27- -02-   -27- -02-    |||||
  -26- -01-   -26- -01-    |||||
                            J6
on J4 and J5:  pin 1 is the bottom right pin in the 25row by 2 columns

On J4, there are 23 free pins
On J5, there are 23 free pins

Pins 49 and 50 on J4 are REFCLK+ and REFCLK- driven from the FPGA
on B12.  The firmware below drives this pin (B12) from the DCM through
a TTL->LVPECL converter (U2A).  The origin of this clock is the 40MHz
clock from the TTCrx

We need an 80MHz clock to latch things.  The DCM produces a 80MHz clock
  which is 2x the TTC clock.  Drive this onto J5/pin1 from U3 (CK80)

Data from each HTR Xilinx:

*/  

`include "define.v"
module SandwichTop 
 ( 
 input  CLKp, CLKn, // 40MHz differential RX_CLK
 input  BCOp, BCOn, // RX_BC0 common to all SLB boards
 input  SLB_ID,
 input  TTC_CK40Des1, TTC_BRCST_STRB1, //TTC_BCR,
 input  SCLK,       // 80 MHz Clock from onboard crystal
 input  LOC_nCS,  LocWRITE_b, 
 input  [7:2] TTC_BRCST,
 input [23:0] TOP_D, BOT_D,		// data from HTR motherboard
 input [11:0] TOPbus, BOTbus,    //               "
 input  [`LB_Addr_Size-1:0]  LocAdd,  // The Address bus.
 inout  [`REG_WIDTH-1:0]     LocBus,  // The bidir Data bus 
 output SLB_CFGDONE,
 output [22:0] SPY4,		// ouput onto J4  see http://www.physics.umd.edu/hep/HTR/luminosity/uwspy.html
 output [22:0] SPY5,		// ouput onto J5                            ditto

 output LVDS1n, LVDS1p, LVDS2n, LVDS2p,// to MDR14 J10
 output REFCLK, TRSTN,  LED_CONTROL          );

//
// declarations...
//
reg[26:0] clk_scaled;
wire  VME_Hard_rst, VME_Soft_rst, VME_Start, VME_Stop, VME_LVDSout,
       run_status, BunchEr, GateTTCcommands, Hard_rst, Soft_rst;
wire resetn = ~Hard_rst;
wire [15:0] TTC_test0, TTC_test1, TTC_test2, TTC_test3,
             TTC_test4, TTC_test5, TTC_test6, TTC_test7,
             TTC_test8, TTC_test9, TTC_testA, TTC_testB,
             TTC_testC, TTC_testD, TTC_testE, TTC_testF;

wire [8:1]  TPSpyFifoEmpty, TPSpyFifoFull, TPSpyFifoRd,
             InstantCntErr,  InstantR1sErr;
wire [15:0] AccCntErr, AccR1sErr;  
wire [35:0]  TPSpyFIFO1_out, TPSpyFIFO2_out, TPSpyFIFO3_out;
reg VME_LVDSout_D, LVDSout1, LVDSout2;
wire [7:0] LastTTCcommand;
reg [23:0] rTOP_D, rBOT_D;
reg [11:0] rTOPbus, rBOTbus;
reg [15:0] RxClkCnt, RxBC0Cnt, TClkCnt;
wire[7:0] mux_out;

wire[1:0] ch1,ch2;
wire[3:0] mch1;
wire[5:0] sum0, sum1, sum2, sum3;
wire hist_localbus_read;
wire[11:0] hist_localbus_bin;
wire hist_zero;
wire[11:0] bucket;  // reset when "BC0" is asserted
wire readData;
wire [3:0] hist_localbus_select;
wire HistogramMode;
wire[2:0] lumState;
wire latch_all, hdone;
//  ********************************************************************
//      ********************  CLOCK SECTION    *****************
//            *********************************************

wire   RxClk_Buf, SCLK_buf;
wire   Tclk0, Tclk180, TTC_CK, Tclk2x_dll, TTC_CK2x, TTC_DLL_lock;
reg    [2:0] UnlockCnt; // counter for proper reset of DLL 
//
// Instantiate the Differential Global Input Buffer 
//
IBUFGDS_LVPECL_25 difclkbuf (.I(CLKp), .IB(CLKn), .O(RxClk_Buf));
IBUFG sclkbuf ( .I(SCLK), .O(SCLK_buf) );
//                  
// Generate a DLL reset (3-ticks) after FPGA is configured, see:
// www.xilinx.com/publications/xcellonline/xcell_51/xc_particle51.htm
//
  FDS flop1 (.D(1'b0), .C(SCLK_buf), .Q(out1), .S(1'b0)); 
  FD  flop2 (.D(out1), .C(SCLK_buf), .Q(out2)); 
  FD  flop3 (.D(out2), .C(SCLK_buf), .Q(out3)); 
  FD  flop4 (.D(out3), .C(SCLK_buf), .Q(out4)); 
assign ConfigRst = (out2 | out3 | out4); 
//
// instantiate the DCM
//
DCM TTC_DCM
   (.CLKIN(TTC_CK40Des1), .CLKFB(TTC_CK),  .RST(ConfigRst),  
    .CLK0(Tclk0),   .CLK90(),  .CLK180(Tclk180),  .CLK270(),	
    .CLK2X(),  .CLK2X180(Tclk2x_dll),
    .CLKDV(),   .CLKFX(),   .CLKFX180(),
    .LOCKED(TTC_DLL_lock),  .STATUS(),
    .DSSEN(1'b0),
    .PSCLK(1'b0),  .PSEN(1'b0),  .PSINCDEC(1'b0), .PSDONE() );

BUFG TCK40BUF (.I(Tclk0), .O(TTC_CK));
BUFG TCK40BUF180 (.I(Tclk180), .O(TTC_CK180));  
BUFG TCK2XBUF (.I(Tclk2x_dll), .O(TTC_CK2x)); 

//
//register data since it is arriving at 80MHz
//
always @ (posedge TTC_CK2x) begin 
  rTOP_D <= TOP_D;
  rBOT_D <= BOT_D;
  rTOPbus <= TOPbus;
  rBOTbus <= BOTbus;
  end
//
// now form the sums each channel
//
wire [7:0] sdebug;
assign mch1 = sdebug[7:4];
assign ch1 = sdebug[3:2];
assign ch2 = sdebug[1:0];
wire rframe = rTOPbus[11];  
wire rTTC_TestEnable = rTOPbus[10];
channelSums TOPsum(.clk(TTC_CK180),.clk2x(TTC_CK2x),.frame(rframe),.data(rTOP_D),
	.sum0(sum0),.sum1(sum1),.sum2(sum2),.sum3(sum3),
	.debug(sdebug));
//
// use TTC_BRCST_STRB1 as our BC0...for now...
//
wire bc0_strobe = TTC_BRCST_STRB1;
//
// histograms go here...to start filling, issue the "START" command and set the CSR
// bit to enable histogram filling. filling will commence after the next "BC0"
//	note:  when reading, all 4 histograms are read
wire startHistograms = HistogramMode && run_status;
wire [15:0] rv1,rv2,rv3,rv4;
wire [1:0] rs1,rs2,rs3,rs4;
wire rw1,rw2,rw3,rw4;
wire [3:0] sel0,sel1,sel2,sel3;
wire[9:0] addr0,addr1;
wire[1:0] hstate0,hstate1;
wire[3:0] hLatch;
wire[15:0] latch_h0, latch_h1,latch_h2,latch_h3;
wire[11:0] freeBucket;
wire szero;
histLUM HFLUM(
	// inputs:
	.clk(TTC_CK180),.clkn(TTC_CK),.resetn(resetn),
	.fEnable(startHistograms),.BC0(bc0_strobe),.zero(hist_zero),
	.rEnable(hist_localbus_read),.rAddress(hist_localbus_bin),.rWhich(hist_localbus_select),
	.sum0(sum0),.sum1(sum1),.sum2(sum2),.sum3(sum3),
	// outputs:
	.goRead(readData),.bucket(bucket),.histDone(HistogramsDone),.latch_all(latch_all),.latchEach(hLatch),
	.hData0(latch_h0),.hData1(latch_h1),.hData2(latch_h2),.hData3(latch_h3),.orbitDone(hdone),
	.enable_ram_write(enable_ram_write),.enable_ram_read(enable_ram_read),.state(lumState),
	// debugging:
	.rv1(rv1),.rv2(rv2),.rv3(rv3),.rv4(rv4),
	.rs1(rs1),.rs2(rs2),.rs3(rs3),.rs4(rs4),
	.rw1(rw1),.rw2(rw2),.rw3(rw3),.rw4(rw4),
	.sel0(sel0),.sel1(sel1),.sel2(sel2),.sel3(sel3),
	.addrH0(addr0),.addrH1(addr1),.h0State(hstate0),.h1State(hstate1),.freeBucket(freeBucket),
	.szero(szero)
	);
//
// spy output onto wisconsin receiver
//	see http://www.physics.umd.edu/hep/HTR/luminosity/uwspy.html
// there are 23 pins available per wisconsin connectors J4 and J5
// SPY4 has is hooked up to various bits of J4, and likewise for SPY5 and J5
// i load up SPY4 and SPY5 in some kind of random fashion due to how I hook
// up the agilent.
//
//assign SPY4 = clk_scaled[22:0];
//assign SPY5 = clk_scaled[22:0];
reg[15:0] spyj5, spyj4;
//wire[15:0] spyj5, spyj4;
assign SPY4 = {5'h0,spyj4[15:2],1'b0,spyj4[1],1'b0,spyj4[0]};
assign SPY5 = {5'h0,spyj5[15:10],1'b0,spyj5[9:8],1'b0,spyj5[7:0]};
//assign spyj4 = clk_scaled[15:0];
//assign spyj5 = clk_scaled[15:0];

wire[11:0] temp = {enable_ram_write,enable_ram_read,lumState[2:0],hist_localbus_read,
		                latch_all,bc0_strobe,hdone,startHistograms,szero,HistogramMode};
always @ (mux_out or ch1 or ch2 or mch1 or sum0 or sum1 or sum2 or sum3 or rTOP_D or TOP_D or bucket
  or TTC_CK180 or TTC_CK2x or TTC_BRCST_STRB1 or rframe) begin
	case (mux_out[3:0])
		0: begin
         spyj5[15:12] = {TTC_CK180, TTC_CK2x, TTC_BRCST_STRB1, rframe};
			spyj5[11:0] = {ch2,ch1,mch1,TOP_D[1:0],rTOP_D[1:0]};
			spyj4 = {spyj5[15:12],sum1,sum0};
			end
		1: begin
         spyj5[15:12] = {TTC_CK180, TTC_CK2x, TTC_BRCST_STRB1, rframe};
			spyj5[11:0] = {ch2,ch1,mch1,TOP_D[1:0],rTOP_D[1:0]};
			spyj4 = {spyj5[15:12],sum3,sum2};
			end
	   2: begin
         spyj5[15:12] = {TTC_CK180, TTC_CK2x, TTC_BRCST_STRB1, rframe};
		   spyj5[11:0] = {ch2,ch1,mch1,TOP_D[1:0],rTOP_D[1:0]};
			spyj4 = {sum1[3:0],bucket};
			end
		3: begin
         spyj5[15:12] = {TTC_CK180, TTC_CK2x, TTC_BRCST_STRB1, rframe};
			spyj5[11:0] = temp;
			spyj4 = {mch1,freeBucket};
//			spyj4 = {mch1,bucket};
////		   spyj4 = {sum0[3:0],bucket};
		   end
	   4: begin
         spyj5[15:12] = {TTC_CK180, TTC_CK2x, TTC_BRCST_STRB1, rframe};
		   spyj5[11:0] = rTOP_D[11:0];
			spyj4 = {mch1,rTOP_D[23:12]};
			end
	   5: begin
         spyj5[15:12] = {TTC_CK180, TTC_CK2x, TTC_BRCST_STRB1, rframe};
		   spyj5[11:0] = temp;
			spyj4 = {sel0[3:0],hstate0[1:0],addr0[9:0]};
			end
	   6: begin
         spyj5[15:12] = {TTC_CK180, TTC_CK2x, TTC_BRCST_STRB1, rframe};
		   spyj5[11:0] = temp;
			spyj4 = {sel1[3:0],hstate1[1:0],addr1[9:0]};
			end
		7: begin
         spyj5[15:12] = {TTC_CK180, TTC_CK2x, TTC_BRCST_STRB1, rframe};
		   spyj5[11:0] = temp;
		   spyj4 = {sel1[3:0],rw4,rw3,rw2,rw1,
				rs4[1:0],rs3[1:0],rs2[1:0],rs1[1:0]};
		   end
		8: begin
         spyj5[15:12] = {TTC_CK180, TTC_CK2x, TTC_BRCST_STRB1, startHistograms};
		   spyj5[11:0] = bucket;
		   spyj4 = rv1;
		   end
		9: begin
         spyj5[15:12] = {TTC_CK180, TTC_CK2x, TTC_BRCST_STRB1, startHistograms};
		   spyj5[11:0] = bucket;
		   spyj4 = rv2;
		   end
		'd10: begin
         spyj5[15:12] = {TTC_CK180, TTC_CK2x, TTC_BRCST_STRB1, startHistograms};
		   spyj5[11:0] = bucket;
		   spyj4 = rv3;
		   end
		'd11: begin
         spyj5[15:12] = {TTC_CK180, TTC_CK2x, TTC_BRCST_STRB1, startHistograms};
		   spyj5[11:0] = bucket;
		   spyj4 = rv4;
		   end
		'd12: begin
         spyj5[15:12] = {TTC_CK180, TTC_CK2x, TTC_BRCST_STRB1, rframe};
		   spyj5[11:0] = {enable_ram_write,enable_ram_read,lumState[2:0],hist_localbus_read,
		                latch_all,bc0_strobe,hLatch[3:0]};
			spyj4 = latch_h1;
		   end
   endcase
end

always @ (posedge TTC_CK2x) clk_scaled <= #1 clk_scaled + 1;
assign LED_CONTROL = clk_scaled[25];

//LED_Sine breathing(.clock(SCLK_buf), .nReset(1'b1), .led(LED_CONTROL));
 assign REFCLK = TTC_CK;
 assign SLB_CFGDONE = SLB_ID; // in case ECAL TCC uses this board
 assign TRSTN = 1'b1; // Wisc. Rec_Mezz JTAG reset.
 IBUFDS_LVPECL_25 RxBC0buf (.I(BCOp), .IB(BCOn), .O(RxBC0));
 OBUFDS_LVDS_25 pair1 (.I(LVDSout1), .O(LVDS1p), .OB(LVDS1n));
 OBUFDS_LVDS_25 pair2 (.I(LVDSout2), .O(LVDS2p), .OB(LVDS2n));
//
// decode the TTC to keep track (just for checking)
//
decodeTTC TTCdecode (
 .TTC_CK40Des1(TTC_CK),
 .reset(VME_Hard_rst), 
 .TTC_Brcst(TTC_BRCST),
 .TTC_BrcstStr(TTC_BRCST_STRB1),   
 .TTC_test0(TTC_test0), .TTC_test1(TTC_test1),
 .TTC_test2(TTC_test2), .TTC_test3(TTC_test3),
 .TTC_test4(TTC_test4), .TTC_test5(TTC_test5),
 .TTC_test6(TTC_test6), .TTC_test7(TTC_test7),
 .TTC_test8(TTC_test8), .TTC_test9(TTC_test9),
 .TTC_testA(TTC_testA), .TTC_testB(TTC_testB),
 .TTC_testC(TTC_testC), .TTC_testD(TTC_testD),
 .TTC_testE(TTC_testE), .TTC_testF(TTC_testF),
 .LastTTCcommand(LastTTCcommand)
 );

always @ (posedge TTC_CK) begin
  LVDSout1 <= VME_LVDSout;
  LVDSout2 <= VME_LVDSout;
  end

//
// put the entire 36 bits from each Xilinx (HTR TOP and BOT) into a fifo
// which is 512 deep and 36 bits wide (RAMB16_S36_S36).  however, the
// readout data path is only 16 bits wide.  what to do?   we'll have to 
// read it 16 bits at a time!	 therefore we'll need a different register
// for the 3 reads (3x16 > 36 bits).  all this will be done inside the
// localbusslave routine
//
fifo_ic_v2  TPSpyFIFO1 
 ( .fifo_gsr_in(Hard_rst),       .write_data_in({rTOPbus, rTOP_D}), 
   .write_clock_in(TTC_CK2x),        .write_enable_in(run_status),
   .read_clock_in(TTC_CK),           .read_enable_in(TPSpyFifoRd[1]),
   .read_data_out({TPSpyFIFO1_out}), .fifostatus_out(),
   .full_out(TPSpyFifoFull[1]),      .empty_out(TPSpyFifoEmpty[1])  );

fifo_ic_v2  TPSpyFIFO2 
 ( .fifo_gsr_in(Hard_rst),       .write_data_in({rBOTbus, rBOT_D}), 
   .write_clock_in(TTC_CK2x),        .write_enable_in(run_status),
   .read_clock_in(TTC_CK),           .read_enable_in(TPSpyFifoRd[2]),
   .read_data_out({TPSpyFIFO2_out}), .fifostatus_out(),
   .full_out(TPSpyFifoFull[2]),      .empty_out(TPSpyFifoEmpty[2])  );
												 
fifo_ic_v2  TPSpyFIFO3 
 ( .fifo_gsr_in(Hard_rst),       .write_data_in({12'b0,sum3,sum2,sum1,sum0}), 
   .write_clock_in(TTC_CK),        .write_enable_in(run_status),
   .read_clock_in(TTC_CK),           .read_enable_in(TPSpyFifoRd[3]),
   .read_data_out({TPSpyFIFO3_out}), .fifostatus_out(),
   .full_out(TPSpyFifoFull[3]),      .empty_out(TPSpyFifoEmpty[3])  );
												 
////////////////////////////////////////////////////////

reg[11:0] count_bucket, latch_bucket;
LocBusSlave LocBusSlave (
    .clk(TTC_CK),      .rst(Hard_rst), 
    .LocAddr(LocAdd),    .LocData(LocBus), 
    .LocCS_b(LOC_nCS),   .LocWRITE_b(LocWRITE_b), 
    .LB_STATE(),   .Din(), .Dout(),      .Tri(), 
    .RxDLL_lock(1'b0), .TTC_DLL_lock(TTC_DLL_lock), .UnlockCnt(UnlockCnt),
    .DLL_status(16'b0),
    .LastTTCcommand(LastTTCcommand), .GateTTCcommands(GateTTCcommands), 
    .VME_Hard_rst(VME_Hard_rst),     .VME_Soft_rst(VME_Soft_rst), 
    .VME_Start(VME_Start),           .VME_Stop(VME_Stop), 
    .VME_LVDSout(VME_LVDSout), 
    .run_status(run_status),         
    .RxClkCnt(RxClkCnt), .RxBC0Cnt(RxBC0Cnt),
    .TClkCnt(TClkCnt), .TxBC0Cnt(16'b0),
    .TPSpyFifoRd(TPSpyFifoRd),
    .TPSpyFifoEmpty(TPSpyFifoEmpty),   .TPSpyFifoFull(TPSpyFifoFull), 
    .TPSpyFIFO1_out(TPSpyFIFO1_out),   .TPSpyFIFO2_out(TPSpyFIFO2_out),
    .TPSpyFIFO3_out(TPSpyFIFO3_out),//   .TPSpyFIFO4_out(TPSpyFIFO4_out),
    .TTC_test0(TTC_test0),       .TTC_test1(TTC_test1), 
    .TTC_test2(TTC_test2),       .TTC_test3(TTC_test3), 
    .TTC_test4(TTC_test4),       .TTC_test5(TTC_test5), 
    .TTC_test6(TTC_test6),       .TTC_test7(TTC_test7), 
    .TTC_test8(TTC_test8),       .TTC_test9(TTC_test9), 
    .TTC_testA(TTC_testA),       .TTC_testB(TTC_testB), 
    .TTC_testC(TTC_testC),       .TTC_testD(TTC_testD), 
    .TTC_testE(TTC_testE),       .TTC_testF(TTC_testF), 
    .BunchEr(BunchEr), .mux_out(mux_out),
	 .latch_bucket(latch_bucket), .count_bucket(count_bucket),
	 .HistogramsDone(HistogramsDone),.HistogramMode(HistogramMode),.HZero(hist_zero),
	 .hist_localbus_read(hist_localbus_read),.hist_localbus_bin(hist_localbus_bin),
	 .hist_localbus_select(hist_localbus_select),
	 .hist0_value(latch_h0),.hist1_value(latch_h1),.hist2_value(latch_h2),.hist3_value(latch_h3)
        );

 commander commander (  
  .clk(SCLK_buf),    .GateTTCcommands(GateTTCcommands),
  .TTC_Hard_rst(1'b0), .TTC_Soft_rst(1'b0),
  .TTC_Start(1'b0), .TTC_Stop(1'b0), 
  .VME_Hard_rst(VME_Hard_rst), .VME_Soft_rst(VME_Soft_rst), 
  .VME_Start(VME_Start), .VME_Stop(VME_Stop), 
  .Hard_rst(Hard_rst),  .Soft_rst(Soft_rst), 
  .run_status(run_status)       );

 always @ (posedge RxClk_Buf) begin 
   if (Soft_rst) begin
      RxClkCnt <= 0;  RxBC0Cnt <= 0;
   end
   else if (run_status) begin
      RxClkCnt <= RxClkCnt + 1;
      if (RxBC0) RxBC0Cnt <= RxBC0Cnt + 1;
   end
 end // always

//
// let's make sure we have the right number of buckets between "bc0"
//
reg[1:0] lhc_state;
parameter[1:0] BC0_WAIT=0, COUNT=1, WAIT_RESET=2;
 always @ (posedge TTC_CK180)
   if (~resetn) begin
	  lhc_state <= BC0_WAIT;
	  count_bucket <= 0; 
	  latch_bucket <= 0;
	end
   else case(lhc_state)
		BC0_WAIT: begin
			count_bucket <= 0;
			latch_bucket <= 0;
			if (run_status && bc0_strobe) lhc_state <= COUNT;
			else lhc_state <= BC0_WAIT;
			end
	   COUNT: begin
			if (bc0_strobe) begin
				latch_bucket <= bucket;
				lhc_state <= WAIT_RESET;
				end
			else begin
		      count_bucket <= count_bucket + 1;
			   lhc_state <= COUNT;
				end
			end
	   WAIT_RESET:
		   if (run_status) lhc_state <= WAIT_RESET;
			else lhc_state <= BC0_WAIT;
   endcase

endmodule